#!/usr/bin/env python
try:
    from    .                   import pftag
except:
    from pftag                  import pftag

from    pathlib                 import Path
from    argparse                import ArgumentParser,                  \
                                       Namespace,                       \
                                       ArgumentDefaultsHelpFormatter,   \
                                       RawTextHelpFormatter

from importlib.metadata import Distribution

__pkg       = Distribution.from_name(__package__)
__version__ = __pkg.version

import  os, sys, json
import  pudb
from    pudb.remote             import set_trace
from    pftag.pftag             import parser_setup, parser_interpret, parser_JSONinterpret


DISPLAY_TITLE = r"""

        __ _
       / _| |
 _ __ | |_| |_ __ _  __ _
| '_ \|  _| __/ _` |/ _` |
| |_) | | | || (_| | (_| |
| .__/|_|  \__\__,_|\__, |
| |                  __/ |
|_|                 |___/

"""

str_desc: str                =  DISPLAY_TITLE + """

                        -- version """ + __version__ + """ --

        An in-place string token lookup and optional function processor. This
        package provides both a CLI client as well as a python module.


"""

package_CLIself:str         = """
        --tag <tagString>                                                       \\
        [--tagMarker <mark>]                                                    \\
        [--funcMarker <mark>]                                                   \\
        [--funcArgMarker <mark>]                                                \\
        [--funcSep <mark>]                                                      \\
        [--test]                                                                \\
        [--inputdir <inputdir>]                                                 \\
        [--outputdir <outputdir>]                                               \\
        [--man]                                                                 \\
        [--verbosity <level>]                                                   \\
        [--debug]                                                               \\
        [--debugTermsize <cols,rows>]                                           \\
        [--debugHost <0.0.0.0>]                                                 \\
        [--debugPort <7900>]"""

package_CLIsynpsisArgs:str  = """
    ARGUMENTS

        --tag <tagString>
        The tag string to process.

        [--tagMarker <mark>]
        The marker string that identifies a tag (default "%")

        [--funcMarker <mark>]
        The marker string that pre- and post marks a function (default "_").

        [--funcArgMarker <mark>]
        The marker string between function arguments and also between arg list
        and function (default "|").

        [--funcSep <mark>]
        The marker string separating successive function/argument constructs
        (default ",").

        [--test]
        If specified, run a small internal test on multi-logger calls.

        [--pftelUser <user>] ("chris")
        The name of the pftel user. Reserved for future use.

        [--inputdir <inputdir>]
        An optional input directory specifier. Reserverd for future use.

        [--outputdir <outputdir>]
        An optional output directory specifier. Reserved for future use.

        [--man]
        If specified, show this help page and quit.

        [--verbosity <level>]
        Set the verbosity level. The app is currently chatty at level 0 and level 1
        provides even more information.

        [--debug]
        If specified, toggle internal debugging. This will break at any breakpoints
        specified with 'Env.set_trace()'

        [--debugTermsize <253,62>]
        Debugging is via telnet session. This specifies the <cols>,<rows> size of
        the terminal.

        [--debugHost <0.0.0.0>]
        Debugging is via telnet session. This specifies the host to which to connect.

        [--debugPort <7900>]
        Debugging is via telnet session. This specifies the port on which the telnet
        session is listening.
"""

package_CLIfunctions:str = r"""

    FUNCTIONS

        OVERVIEW
        In addition to performing a lookup on a template string token, this
        package can also process the lookup value in various ways. These
        process functions follow a Reverse Polish Notation (RPN) schema of

            tag func1 func2 func3 ...

        where first the <tag> is looked up, then this lookup is processed by
        <func1>. The result is then processed by <func2>, and so on and
        so forth.

        This RPN approach also mirrors the standard UNIX piping schema.

        A function that is to be applied to a <tag> should be connected
        to the tag with a <funcMarker> string, usually '_'. The final
        function should end with the same <funcMarker>, so

            %tag_func_

        will apply the function called "func" to the tag called "tag".

        Some functions can accept arguments. Arguments are passed to a function
        with a <funcArgMarker> string, typically '|', that also separates
        arguments:

            %tag_func|a1|a2|a3_

        will pass 'a1', 'a2', and 'a3' as parameters to "func".

        Finally, several functions can be chained within the '_'...'_' by
        separating the <func>|<argList> constructs with commas, so

            %tag_func|a1|a2|a3,func2|b1|b2|b3_

        All these special characters (tag marker, function pre- and post,
        arg separation, fand unction separation can be overriden. For instance,
        with a selection of

        --tagMarker "@" --funcMarker "[" --funcArgMarker "," --funcSep "|"

        strings can be specified as

            @tag[func,a1,a2,a3|func2,b1,b2,b3[

        where preference/legibilty is left to the user
"""


package_CLItagsFuncs:str = r"""

        AVAILABLE TAGS AND FUNCTIONS

        The following tags are available:

            %literal   : simply replace the tag with the word 'literal'.
                          This tag is only useful in conjunction with the
                          'echo' function and together they provide a means
                          to inject arbitary text typically for md5 hashing.
            %name      : return the os.name
            %platform  : return the platform.system()
            %release   : return the platform.release()
            %machine   : return the platform.machine()
            %arch      : return the '%s' % platform.architecture()
            %timestamp :  return the a timestamp

        The following tags are available:

        md5|<chars>         : perform an md5hash on the upstream, limit result
                              to <chars> characters

                                eg: "%timestamp_md5|4_"

                              replace the %timestamp in the input string with
                              an md5 hash of 4 chars of the actual timestamp.

        chrplc|<t>|<n>      : replace <t> with <n> in the upstream input.

                                eg: "%timestamp_chrplc|:|-_"

                              replace the %timestamp in the input string with
                              the actual timestamp where all ":" are replaced with
                              "-".

        strmsk|<mask>       : for each '*' in mask pattern use upstream char
                              otherwise replace with <mask> char.

                                eg: "%platform_strmsk|l****_"

                              replace the %platform in the input string with
                              a string that starts with an 'l' and don't change
                              the subsequent 4 characters. If the %platform
                              has more than 4 characters, only return the 5
                              chars as masked.

        dcmname|<s>|<tail> : replace any upstream %VAR with a DICOM formatted
                              name. If <s> is passed, the seed the faker module
                              with <s> (any string) -- this guarantees that calls
                              with that same <s> result in the same name. If
                              <tail> is passed, then append <tail> to the name.

                                eg: %NAME_dcmname_

                             may produce "BROOKS^JOHN". Each call will have
                             a different name. However,

                                %NAME_dcmname|foobar_

                            will always generate "SCHWARTZ^THOMAS". While

                                %NAME_dcmname|foobar|^ANON

                            will generate "SCHWARTZ^THOMAS^ANON"

        echo|<something> :  Best used with the %literal tag for legibility, will
                            replace the tag with <something>. Be careful of commas
                            in the <something>. If they are to be preserved you
                            will need to set --funcSep to something other than a
                            comma.

                                %literal_echo|why-are-we-here?_

                            will replace the %literal with "why-are-we-here".
                            This is most useful when literal data is to obscured
                            in a template. For instance:

                                %literal_echo|Subject12345,md5|5_

                            where say "Subject12345" is privileged information but
                            important to add to the string. In this case, we can
                            add and then hash that literal string. In future,
                            if we know all the privileged strings, we can easily
                            hash and then and lookup in any `pftag` generated
                            strings to resolve which hashes belong to which
                            subjects.

"""

package_CLIexample:str = r"""
    BRIEF EXAMPLES
    Note that some shells (like fish) might require quoting about the tag argument.

        pftag --tag "run-%timestamp-on-%platform.log"
        pftag --tag "run-%timestamp_chrplc|:|-_-on-%platform.log"
        pftag --tag "run-%timestamp_chrplc|:|-,md5|6_-on-%platform.log"

"""

def synopsis(ab_shortOnly = False):
    scriptName = os.path.basename(sys.argv[0])
    shortSynopsis =  '''
    NAME

        pftag

    SYNOPSIS

        pftag                                                               \ '''\
        + package_CLIself + '''

    '''

    description = '''
    DESCRIPTION

        `pftag` is both a script and python module for processing "tag"ed
        strings with optional functions. It is useful when using templated
        strings that need to be processed at "runtime".

    ''' + package_CLIsynpsisArgs + package_CLIfunctions + package_CLItagsFuncs + package_CLIexample
    if ab_shortOnly:
        return shortSynopsis
    else:
        return shortSynopsis + description

parser                  = ArgumentParser(
    description         = '''
A client for pftag
''',
    formatter_class     = RawTextHelpFormatter
)

def earlyExit_check(args) -> int:
    """
    Perform some preliminary checks

    If version or synospis are requested, print these and return
    code for early exit.
    """
    str_help:str = ''
    if args.man or args.synopsis:
        print(str_desc)
        if args.man:
            str_help     = synopsis(False)
        else:
            str_help     = synopsis(True)
        print(str_help)
        return 1
    if args.b_version:
        print("Name:    %s\nVersion: %s" % (__pkg.name, __version__))
        return 1
    if int(args.verbosity) > 1: print(DISPLAY_TITLE)
    return 0

def test_dict() -> dict:
    # Here we create as tagger without the CLI parser, but a JSON
    # object.
    str_tag:str = r'run-%timestamp_chrplc|:|-_-OS-%platform-%NAME_dcmname_-%literal_echo|Subject4,md5|5_.log'
    tagger:pftag.Pftag      = pftag.Pftag({})

    d_tag:dict             = tagger(str_tag)
    print(d_tag['result'])

def main(argv=None) -> int:
    """
    Main method for the programmatical calling the pftag
    module
    """

    # Call the following to collect the logger Namespace
    # and then edit values in the options:Namespace if needed
    parser:ArgumentParser   = parser_setup('A client for logging to a pftel server')
    options:Namespace       = parser_interpret(parser)

    # any reason we should not continue?
    if earlyExit_check(options): return 1

    # set_trace(term_size=(253, 62), host = '0.0.0.0', port = 7900)
    tag:pftag.Pftag         = pftag.Pftag(options)
    d_pftag:dict            = tag.run(options.tag)
    print(d_pftag['result'])

    if options.test: test_dict()

    return 0 if d_pftag['status'] else 2

if __name__ == '__main__':
    sys.exit(main(sys.argv))



#!/usr/bin/env python
__version__ = '1.0.0'

from    pathlib                 import Path

import  os, sys, json, platform, re
import  pudb
from    pudb.remote             import set_trace

from    concurrent.futures      import ThreadPoolExecutor
from    threading               import current_thread

from    datetime                import datetime, timezone

from    typing                  import Any, Callable, Tuple
from    faker                   import  Faker

import  logging
import  hashlib

import  math
from    argparse                import  Namespace, ArgumentParser
from    argparse                import  RawTextHelpFormatter

from    .                       import  data

# Turn off logging for the 'faker' module and create a class instance
# of the object
fakelogger: logging.Logger  = logging.getLogger('faker')
fakelogger.propagate        = False
fake:Faker                  = Faker()


def parser_setup(str_desc) -> ArgumentParser:
    parser:ArgumentParser = ArgumentParser(
                description         = str_desc,
                formatter_class     = RawTextHelpFormatter
            )

    parser.add_argument(
                '--version',
                default = False,
                dest    = 'b_version',
                action  = 'store_true',
                help    = 'print version info'
    )
    parser.add_argument(
                '--man',
                default = False,
                action  = 'store_true',
                help    = 'show a man page'
    )
    parser.add_argument(
                '--osenv',
                default = False,
                action  = 'store_true',
                help    = 'show the base os environment'
    )
    parser.add_argument(
                '--synopsis',
                default = False,
                action  = 'store_true',
                help    = 'show a synopsis'
    )
    parser.add_argument(
                '--test',
                default = False,
                action  = 'store_true',
                help    = 'call an internal test method'
    )
    parser.add_argument(
                '--inputdir',
                default = './',
                help    = 'optional directory specifying extra input-relative data'
    )
    parser.add_argument(
                '--outputdir',
                default = './',
                help    = 'optional directory specifying location of any output data'
    )
    parser.add_argument(
                '--tag',
                default = '',
                help    = 'tag string to process'
    )
    parser.add_argument(
                '--tagMarker',
                default = '%',
                help    = 'the marker string that identifies a tag (default "%")'
    )
    parser.add_argument(
                '--funcMarker',
                default = '_',
                help    = 'the marker string that pre- and post marks a function (default "_")'
    )
    parser.add_argument(
                '--funcArgMarker',
                default = '|',
                help    = 'the marker string between function arguments and also between arg list and function (default "|")'
    )
    parser.add_argument(
                '--funcSep',
                default = ',',
                help    = 'the separation string between successive function/argument constructs (default ",")'
    )
    parser.add_argument(
                '--verbosity',
                default = '0',
                help    = 'verbosity level of app'
    )
    parser.add_argument(
                "--debug",
                help    = "if true, toggle telnet pudb debugging",
                dest    = 'debug',
                action  = 'store_true',
                default = False
    )
    parser.add_argument(
                "--debugTermSize",
                help    = "the terminal 'cols,rows' size for debugging",
                default = '253,62'
    )
    parser.add_argument(
                "--debugPort",
                help    = "the debugging telnet port",
                default = '7900'
    )
    parser.add_argument(
                "--debugHost",
                help    = "the debugging telnet host",
                default = '0.0.0.0'
    )
    return parser

def parser_interpret(parser, *args):
    """
    Interpret the list space of *args, or sys.argv[1:] if
    *args is empty
    """
    if len(args):
        args    = parser.parse_args(*args)
    else:
        args    = parser.parse_args(sys.argv[1:])
    return args

def parser_JSONinterpret(parser, d_JSONargs):
    """
    Interpret a JSON dictionary in lieu of CLI.
    For each <key>:<value> in the d_JSONargs, append to
    list two strings ["--<key>", "<value>"] and then
    argparse.
    """
    l_args  = []
    for k, v in d_JSONargs.items():
        if type(v) == type(True):
            if v: l_args.append('--%s' % k)
            continue
        l_args.append('--%s' % k)
        l_args.append('%s' % v)
    return parser_interpret(parser, l_args)

class Pftag:

    def env_setup(self, options: Namespace) -> bool:
        """
        Setup the environment

        Args:
            options (Namespace):    options passed from the CLI caller
        """
        status  : bool          = True
        options.inputdir        = Path(options.inputdir)
        options.outputdir       = Path(options.outputdir)
        self.env.inputdir       = options.inputdir
        self.env.outputdir      = options.outputdir
        self.env.debug_setup(
                    debug       = options.debug,
                    termsize    = options.debugTermSize,
                    port        = options.debugPort,
                    host        = options.debugHost
        )
        # if not len(options.tag):
        #     self.env.ERROR("The '--tag <tag>' CLI MUST be specified!")
        #     status              = False
        return status

    def __init__(self, options, *args, **kwargs):
        """
        constructor

        Responsible primarily for setting up the client connection
        to the pftel server.

        Possible TODO? How to check _elegantly_ on dead server?
        """
        global Env
        self.env:data.env                   = data.env()

        self.d_tagReserved:dict[str, list[str]]                  = {
            'core'      : ['literal', 'timestamp', 'os', 'platform', 'release', 'name', 'arch']
        }

        # If the <options> is a dictionary, then we interpret this as if
        # it were the CLI (instead of the real CLI) and convert to a
        # namespace
        if type(options) is dict:
            parser:ArgumentParser           = parser_setup('Setup client using dict')
            options:Namespace               = parser_JSONinterpret(parser, options)
        if type(options) is Namespace:
            self.options:Namespace          = options
            self.env.options                = options
            self.envOK:bool                 = True
        if not self.env_setup(options):
            self.env.ERROR("Env setup failure, exiting...")
            self.envOK                  = False
        self.str_tagMarker:str              = options.tagMarker     # '%
        self.str_funcMarker:str             = options.funcMarker    # '_'
        self.str_funcArgSep:str             = options.funcArgMarker # '|'
        self.str_funcSep:str                = options.funcSep       # ','

        self.env_show()

    def env_show(self) -> None:
        """
        Perform some setup

        Args:
            None (internal self)

        Returns:
            None
        """

        if int(self.options.verbosity) < 4: return

        self.self.env.DEBUG("app arguments...", level = 3)
        for k,v in self.options.__dict__.items():
             self.self.env.DEBUG("%25s:  [%s]" % (k, v), level = 3)
        self.self.env.DEBUG("", level = 3)

        if self.options.osenv:
            self.self.env.DEBUG("base environment...")
            for k,v in os.environ.items():
                self.self.env.DEBUG("%25s:  [%s]" % (k, v), level = 3)
            self.self.env.DEBUG("")

    def tag_findDict(self, tag:str) -> list[Tuple]:
        """
        For a given <tag> superstring, determine if any reserved dictionaries
        contain a token within the <tag>. If found, add the dictionary name and
        token to a tuple list, and eventually return a list of tuples of
        (<dict>, <tag>) containing hits.

        For example, imagine that the string to analyze is

                %platform-name-os

        where the %platform is a tag to be substituted. Imagine further that
        'name' and 'os' are valid tags, but note that they do not have the
        leading tag marker, '%', and so should be literals (i.e. unprocessed).

        Give an tag superstring of 'platform-name-os' this method will
        correctly determine that 'platform' is the tag to analyze and also
        return the internal dictionary name containing this tag.

        Args:
            tag (str): the "tag" to search

        Returns:
            list[Tuple]: a list of tuples with the housing dictionary
            and tag within that dictionary
        """
        T:str                   = self.str_tagMarker
        l_tagPossibilities:list = []    # Is the tag in one of the possible dictionaries?
        l_tagHit:list           = []    # Which tag, *exactly*, is sought in the possibilities?
        tagHit:str              = ''    # This is it!
        l_tagDict:list          = []    # A list of tuples of (<dict>, <tag>)
        for d in list(self.d_tagReserved.keys()):
            l_tagPossibilities:list     = [i for i in self.d_tagReserved[d] if i in tag]
            l_tagHit:list               = [i for i in l_tagPossibilities if f'{T}'+i in f'{T}'+tag]
            tagHit:str                  = l_tagHit[0] if len(l_tagHit) else ''
            if tagHit:                  l_tagDict.append( (d, tagHit) )
        return l_tagDict

    def tag_lookupCore(self, tag:str) -> str:
        """Lookup a tag in the "core" dictionary.

        Args:
            tag (str): the tag to lookup

        Returns:
            str: the string lookup
        """
        timenow:Callable[[], str]       = lambda: datetime.now(timezone.utc).astimezone().isoformat()
        lookup:str  = ""
        tag:str     = tag.lower()
        if 'literal'        in tag: lookup  = 'literal'
        if 'name'           in tag: lookup  = os.name
        if 'platform'       in tag: lookup  = platform.system()
        if 'release'        in tag: lookup  = platform.release()
        if 'machine'        in tag: lookup  = platform.machine()
        if 'arch'           in tag: lookup  = '%s-%s' % platform.architecture()
        if 'timestamp'      in tag: lookup  = timenow()
        return lookup

    def tag_lookup(self, tagTuple:tuple) -> str:
        """
        Main dispatching method for looking up a tag across a variety of
        possible use cases (<dictionaryName>).

        Args:
            tagTuple (tuple): a tuple of <dictionaryName> and <tag> to lookup

        Returns:
            str: the lookup value for the <tag>
        """
        lookup:str  = ""
        match tagTuple[0]:
            case 'core':
                lookup  = self.tag_lookupCore(tagTuple[1])
        return lookup

    def tag_process(self, astr:str, *args, **kwargs):
        """
        This method processes a string that contains "%<tag>" tokens in a
        variety of ways. Various %<tags> are understood:

            * %timestamp     - a timestamp
            * %os            - the OS name
            * %platform      - the platform system
            * %release       - the platform release

        For example, an input tag that is specified as the following string:

                    %os-%platform-output.txt

        will be parsed to

                    posix-Linux-6.2.2-zen1-1-zen-output.txt

        It is also possible to apply certain permutations/functions
        to a tag. For example, a function is specified as

                %<tag>_<funcName>|<arg1>|<arg2>...

        which will apply the <funcName> (with <arg1>, <arg2>) to the
        lookup value of <tag>
        """

        def echo(funcArgs:str, strproc:str) -> str:
            """
            Simply replace the tag with an "echo" passed
            in args.

            This is useful to inject a client side string
            into the superstring payload

            Args:
                funcArgs (str): the function_and_args construct
                strproc (str): the string to process

            Returns:
                str: the processed string
            """
            F:str               = self.str_funcMarker
            A:str               = self.str_funcArgSep
            l_args:list[str]    = funcArgs.split(f'{A}')
            value:str           = strproc.split(f'{F}')[0]
            result:str          = l_args[1]

            return result

        def strmsk(funcArgs:str, strproc:str) -> str:
            """
            string mask
            """
            F:str               = self.str_funcMarker
            A:str               = self.str_funcArgSep
            l_args:list[str]    = funcArgs.split(f'{A}')
            value:str           = strproc.split(f'{F}')[0]
            str_msk:str         = l_args[1]
            l_n:list            = []
            result:str          = ""

            for i, j in zip(list(value), list(str_msk)):
                if j == '*':    l_n.append(i)
                else:           l_n.append(j)
            result  = ''.join(l_n)
            return result

        def md5(funcArgs:str, strproc:str) -> str:
            """
            Apply an md5 hash on the <astr>. A single argument
            is accepted, denoting the length of the return hash.

            Args:
                funcArgs (str): the function_and_args construct
                strproc (str): the string to process

            Returns:
                str: the processed string
            """
            F:str               = self.str_funcMarker
            A:str               = self.str_funcArgSep
            l_args:list[str]    = funcArgs.split(f'{A}')
            value:str           = strproc.split(f'{F}')[0]
            chars:str               = ''
            result:str              = hashlib.md5(value.encode('utf-8')).hexdigest()
            if len(l_args) > 1:
                chars       = l_args[1]
                result:str  = result[0:int(chars)]
            return result

        def chrplc(funcArgs:str , strproc:str)-> str:
            """
            Replace characters in <astr>

            Args:
                funcArgs (str): the function_and_args construct
                strproc (str): the string to process

            Returns:
                str: the processed string
            """
            F:str               = self.str_funcMarker
            A:str               = self.str_funcArgSep
            l_args:list[str]    = funcArgs.split(f'{A}')
            value:str           = strproc.split(f'{F}')[0]
            result:str              = re.sub(f'{l_args[1]}', f'{l_args[2]}', value)
            return result

        def convertToNumber (s:str) -> int:
            return int.from_bytes(s.encode(), 'little')

        def convertFromNumber (n:int)-> bytes:
            return n.to_bytes(math.ceil(n.bit_length() / 8), 'little').decode()

        def dcmname(funcArgs:str, strproc) -> str:
            """
            replace <strproc> with name in DICOM conventions

            If passed a string in the "function" arguments, this is used to seed
            the name caller. If the same argument is used for additional calls to
            this sub-function, then the returned name will be identical.

            Args:
                funcArgs (str): the function_and_args construct
                strproc (str): the string to process

            Returns:
                str: the processed string
            """
            # pudb.set_trace()
            F:str               = self.str_funcMarker
            A:str               = self.str_funcArgSep
            l_args:list[str]    = funcArgs.split(f'{A}')
            value:str           = strproc.split(f'{F}')[0]
            result:str          = ""
            if len(l_args) >  1:
                randSeed:int    = convertToNumber(l_args[1])
                Faker.seed(randSeed)
            str_firstLast:str   = fake.name()
            l_firstLast:list    = str_firstLast.split()
            str_first:str       = l_firstLast[0]
            str_last:str        = l_firstLast[1]
            result              = '%s^%s' % (str_last.upper(), str_first.upper())
            if len(l_args) > 2:
                result += l_args[2]
            return result

        str_replace:str         = ''        # the lookup/processed tag value
        l_tags:list             = []        # the input string split by '%'
        func:str                = ''        # the function to apply
        tag:str                 = ''        # the tag in the funcTag combo
        result:str              = ''        # result of any applied function
        prevResult:str          = ''        # previous function stack result
        tagLookup:str           = ''        # the lookup value for the tag
        d_ret:dict[str, Any]    = {
            'status':           False,
            'funcApplied':      [],
            'result':           ""
        }
        T:str                   = self.str_tagMarker
        F:str                   = self.str_funcMarker
        S:str                   = self.str_funcSep

        if not f'{T}' in astr: return d_ret
        d_ret['status']     = True
        l_tags:list         = astr.split(f'{T}')[1:]
        for tagFunc in l_tags:
            l_tagFunc:list  = tagFunc.split(f'{F}')
            if len(l_tagFunc) > 1:  tag, func = (l_tagFunc[0], l_tagFunc[1])
            else:                   tag, func = (l_tagFunc[0], "")
            lt_hit:list[Tuple]  = self.tag_findDict(tag)
            tagLookup           = tag if not lt_hit else self.tag_lookup(lt_hit[0])
            if func:
                for f in func.split(f'{S}'):
                    tagFunc         = tagFunc.replace(f'{tag}', tagLookup)
                    b_funcDo:bool   = False
                    if 'echo'       in f: result  = echo(f, tagFunc)        ; b_funcDo  = True
                    if 'md5'        in f: result  = md5(f, tagFunc)         ; b_funcDo  = True
                    if 'strmsk'     in f: result  = strmsk(f, tagFunc)      ; b_funcDo  = True
                    if 'chrplc'     in f: result  = chrplc(f, tagFunc)      ; b_funcDo  = True
                    if 'dcmname'    in f: result  = dcmname(f, str_replace) ; b_funcDo  = True
                    d_ret['funcApplied'].append(b_funcDo)
                    if b_funcDo:
                        if f'{T}{tag}{F}{f},'       in astr:    astr:str    = astr.replace(f'{T}{tag}{F}{f},', f'{result}{F}')
                        if f'{T}{tag}{F}{f}{F}'     in astr:    astr:str    = astr.replace(f'{T}{tag}{F}{f}{F}', result)
                        if f'{prevResult}{F}{f},'   in astr:    astr:str    = astr.replace(f'{prevResult}{F}{f},', f'{result}{F}')
                        if f'{prevResult}{F}{f}{F}' in astr:    astr:str    = astr.replace(f'{prevResult}{F}{f}{F}', result)
                        if f'{tagLookup}{F}{f},'    in tagFunc: tagFunc:str = tagFunc.replace(f'{tagLookup}{F}{f},', f'{result}{F}')
                        if f'{tagLookup}{F}{f}{F}'  in tagFunc: tagFunc:str = tagFunc.replace(f'{tagLookup}{F}{f}{F}', result)
                        if f'{prevResult}{F}{f},'   in tagFunc: tagFunc:str = tagFunc.replace(f'{prevResult}{F}{f},', f'{result}{F}')
                        if f'{prevResult}{F}{f}{F}' in tagFunc: tagFunc:str = tagFunc.replace(f'{prevResult}{F}{f}{F}', result)
                    prevResult      = result
            elif lt_hit:
                if f'{T}{lt_hit[0][1]}' in astr:
                    astr:str    = astr.replace(f'{T}{lt_hit[0][1]}', tagLookup)
        d_ret['result'] = astr
        return d_ret

    def run(self, tag:str) -> dict:
        """Main "run" method

        Args:
            tag (str): the tag string to process

        Returns:
            dict: a status and result dictionary
        """
        b_status:bool       = False
        if not self.envOK:  return {'status': b_status}

        result:dict         = self.tag_process(tag)
        return result

    def __call__(self, tag:str, *args: Any, **kwds: Any) -> dict:
        self.options.tag = tag
        return self.run(tag)